"""
Python code for neural network potentials
using the symmetry functions recently developed by
Behler and Parrinello (Phys. Rev. Lett., 2007, 98, 146401;
Phys. Chem. Chem. Phys., 2011, 13, 17930-17955;
Phys. Rev. B, 2012, 86, 079914) as inputs.

"""

######################################################################

import numpy as np
from scipy.optimize import fmin_bfgs as optimizer
import time
from ase.calculators.neighborlist import NeighborList
from ase.calculators.calculator import Calculator
from ase.data import atomic_numbers
from ase.parallel import paropen
from collections import OrderedDict
from ase import io
import os
import json
import warnings
import multiprocessing as mp
try:
    from .. import fmodules  # version 3 of fmodules
    fmodules_version = 3
    wrong_version = fmodules.check_version(version=fmodules_version)
    if wrong_version:
        raise RuntimeError('Fortran part is not updated. Please recompile'
                           ' with f2py as described in the README. '
                           'Correct version is %i.' % fmodules_version)
except ImportError:
    fmodules = None
    warnings.warn('Not using the Fortran part of the code. '
                  'Some parts may be very slow! Compile as described '
                  'in the README for improved performance.')

##############################################################################


class BPNeural(Calculator):

    """Behler-Parrinello Neural Network Potential Calculator.
    Inputs:
        json: string
            output JSON file, or input file if JSON already exists
        cutoff: float (default 6.5 Angstroms)
                  Radius above which neighbor interactions are ignored.
        Gs: dictionary of symbols and lists of dictionaries for making
        symmetry functions. Either auto-genetrated, or given in the following
        form: Gs = {"O": [..., ...], "Au": [..., ...]} where
        ... = {"type":"G2", "element":"O", "eta":0.0009} or
        ... = {"type":"G4", "elements":["O", "Au"], "eta":0.0001,
        "gamma":0.1, "zeta":1.0}
        hiddenlayers: dictionary of chemical element symbols and architectures
                    of their corresponding hidden layers of the conventional
                    neural network (NN). Note that for each atom, number of
                    nodes in the input layer is always equal to the number of
                    symmetry functions (Gs), and the number of nodes in the
                    output layer is always one. For example,
                    hiddenlayers = {"O":(3,5), "Au":(5,6)} means that for "O"
                    we have two hidden layers, the first one with three nodes
                    and the second one having five nodes.
        activation: string to assign the type of activation funtion. "linear"
                    refers to linear function, "tanh" refers to tanh
                    function, and "sigmoid" refers to sigmoid function.
        label: string
            Name used for all files.
        weights: dictionary of symbols and dictionaries of arrays of weights;
                 each symbol and related dictionary corresponds to one chemical
                   element and one conventional neural network.  The weights
                   dictionary for the above example is
                   weights = {"O": {1: np.array(149,3), 2: np.array(4,5),
                   3: np.array(6,1)}, "Au": {1: np.array(148,5),
                   2: np.array(6,6), 3:np.array(7,1)}. The arrays are set up to
                   connect node i in the previous layer with node j in the
                   current layer with indices w[i,j]. There are n+1 rows
                   (i values), with the last row being the bias, and m columns
                   (j values). If weights is not given, the arrays will be
                   randomly generated from values between -0.5 and 0.5.
        scalings: dictionary of parameters for slope and intercept for each
                    element. They are used in order to remove the final value
                    from the range that the activation function would
                    otherwise be locked in. For example,
                      scalings={"Pd":{"slope":2, "intercept":1},
                                "O":{"slope":3, "intercept":4}}
        fp_label: string
            Name used for saving fingerprint in JSON format. If no name is
            given, it is turned off by default, but if a string is assigned, it
            will automatically turn on to save fingerprints.  Saving
            fingerprints can speed up the calculator, particularly in cases
            where images are added sequentially.
        extrapolate: boolean
            If True, allows for extrapolation, if none, doesn't allow.
    Output:
        energy: float
        forces: matrix of floats

    **NOTE: The dimensions of the weight matrix should be consistent with the
            hiddenlayers.**
    """
    implemented_properties = ['energy', 'forces']

    ##########################################################################

    def __init__(self, json=None, cutoff=6.5, Gs=None, hiddenlayers=None,
                 activation='tanh', label=None, weights=None, scalings=None,
                 fp_label=None, extrapolate=True):
        self._cutoff = cutoff
        self._Gs = Gs
        self._hiddenlayers = hiddenlayers
        self._activation = activation
        self._weights = weights
        self._scalings = scalings
        self._label = label
        self._fp_label = fp_label
        self._extrapolate = extrapolate

        # Reading parameters fron json if any:
        if json is not None:
            json_file = paropen(json, 'rb')
            self._cutoff, self._activation, self._hiddenlayers, self._label, \
                self._Gs, self._fingerprint_range, self._weights, \
                self._scalings = read_parameters_json(json_file)

        if label is not None:
            self._label = label

        # Checking that the activation function is given correctly:
        string = 'Unknown activation function; activation must be \
                    either "linear", "tanh", or "sigmoid".'
        if self._activation != 'linear' and self._activation != 'tanh' \
                and self._activation != 'sigmoid':
            raise NotImplementedError(string)

        # Checking the compatibility of the forms of Gs, hiddenlayers and
        # weights:
        string1 = 'Gs and weights are not compatible.'
        string2 = 'hiddenlayers and weights are not compatible.'
        if self._hiddenlayers is not None:
            self._elements = sorted(self._hiddenlayers.keys())
            for element in self._elements:
                if self._weights is not None:
                    if self._Gs is not None:
                        if np.shape(self._weights[element][1])[0] != \
                                len(self._Gs[element]) + 1:
                            raise RuntimeError(string1)
                    if isinstance(self._hiddenlayers[element], int):
                        if np.shape(self._weights[element][1])[1] != \
                                self._hiddenlayers[element]:
                            raise RuntimeError(string2)
                    else:
                        if np.shape(self._weights[element][1])[1] != \
                                self._hiddenlayers[element][0]:
                            raise RuntimeError(string2)
                        for i in \
                                range(2, len(self._hiddenlayers[element]) + 1):
                            if np.shape(self._weights[element][i])[0] != \
                                self._hiddenlayers[element][i - 2] + 1 or \
                                np.shape(self._weights[element][i])[1] != \
                                    self._hiddenlayers[element][i - 1]:
                                raise RuntimeError(string2)

        Calculator.__init__(self, self._label)

    #########################################################################

    def initialize(self, atoms):
        self.par = {}
        self.rc = 0.0
        self.numbers = atoms.get_atomic_numbers()
        self.forces = np.empty((len(atoms), 3))
        self.nl = NeighborList([0.5 * self.rc + 0.25] * len(atoms),
                               self_interaction=False)

    #########################################################################

    def calculate(self, atoms, properties, changes):
        """ Calculation of the energy of the system and forces of all atoms."""

        cf = CalculateFingerprints(self._cutoff, self._Gs, atoms)
        ff = FeedForward(self._hiddenlayers, self._weights,
                         self._scalings, self._activation)

        if not self._extrapolate:
            if compare_train_test_fingerprints(
                    cf,
                    atoms,
                    self._fingerprint_range) == 1:
                raise ExtrapolateError('Trying to extrapolate, which'
                                       ' is not allowed. Change to '
                                       'extrapolate=True if this is'
                                       ' desired.')

        if self._weights is None:
            raise RuntimeError("Calculator not trained; can't return "
                               'properties.')

        if 'numbers' in changes:
            self.initialize(atoms)

        self.nl.update(atoms)

        if properties == ['energy']:
            self.energy = 0.0
            for atom in atoms:
                index = atom.index
                symbol = atom.symbol
                indexfp = cf.index_fingerprint(symbol, index)
                oo, Ei = ff.get_output(indexfp, symbol)
                self.energy += Ei
                self.results['energy'] = float(self.energy)

        if properties == ['forces']:
            self.forces[:] = 0.0
            gf = GetForces(self._cutoff, self._Gs, atoms,
                           self._hiddenlayers, self._weights,
                           self._scalings, self._activation)
            for atom in atoms:
                index = atom.index
                self.forces[index][0] = gf.calculate_force(index, 0)
                self.forces[index][1] = gf.calculate_force(index, 1)
                self.forces[index][2] = gf.calculate_force(index, 2)
            self.results['forces'] = self.forces

    #########################################################################

    def train(
            self,
            images,
            train_forces=None,
            converge=0.005,
            save_fingerprint_derivatives=None):
        """Fits a parameter set to the data, using "fmin_bfgs" optimizer,
        that takes as input a cost function to reduce and an initial guess of
        the parameters and returns an optimized parameter set. The convergence
        criteria is root-mean square of energy fit (eV) per atom.
        Inputs:
            images: list of ASE atoms objects
                    List of objects with positions, symbols, and energies in
                    ASE form. this is the training set of data. This can
                    also be the path to an ASE trajectory (.traj) or
                    database (.db) file.  Energies can be obtained from any
                    reference, e.g. DFT calculations.
            train_forces: if "True", forces will be trained as well
            converge: threshold rmse/image/atom at which simulation is
                    converged
            save_fingerprint_derivatives: if "True", fingerprint derivatives will
            be saved during force training
        """

        if train_forces:
            warnings.warn('Training forces will be slow!')

        if isinstance(images, str):
            extension = os.path.splitext(images)[1]
            if extension == '.traj':
                images = io.PickleTrajectory(images, 'r')
            elif extension == '.db':
                images = io.read(images)

        # Images is converted to dictionary form
        dict_images = {}
        for i in range(len(images)):
            dict_images[i] = images[i]
        images = dict_images.copy()

        # If hiddenlayers is not given, list of elemets and hidden layers are
        # now generated
        if not self._hiddenlayers:
            elements = []
            for image_no in images.keys():
                for atom in images[image_no]:
                    if atom.symbol not in elements:
                        elements.append(atom.symbol)
            self._elements = sorted(elements)
            self._hiddenlayers = make_hiddenlayer_structure(self._elements)

        # If Gs is not given, generates symmetry functions
        if not self._Gs:
            self._Gs = make_symmetry_functions(self._elements, None)

        # If weights are not given, generates random weights
        if not self._weights:
            self._weights = make_weight_matrices(self._Gs, self._elements,
                                                 self._hiddenlayers,
                                                 self._activation)
        # If scalings are not given, generates random scalings
        if not self._scalings:
            self._scalings = make_scalings_matrices(self._elements, images,
                                                    self._activation)
        # Fingerprints are calculated and saved
        self._sfp = SavedFingerprint(
            self._cutoff,
            self._elements,
            self._Gs,
            images,
            self._label,
            self._fp_label,
            train_forces,
            save_fingerprint_derivatives)

        self._fingerprint_range = self._sfp._fingerprint_range

        ravel = ParamRavel(self._hiddenlayers, self._weights, self._scalings)

        costfxn = CostFxnandDer(
            self._cutoff,
            self._Gs,
            self._hiddenlayers,
            ravel,
            images,
            self._activation,
            self._label,
            converge,
            self._sfp,
            self._fp_label,
            self._fingerprint_range)

        vector = ravel.to_vector(self._weights, self._scalings)
        try:
            optimizer(f=costfxn.f, x0=vector, fprime=costfxn.fprime,
                      gtol=10. ** -500.)

        except ConvergedException:
            pass

        vector = ravel.to_vector(costfxn._weights, costfxn._scalings)
        filename = make_filename(self._label, 'trained-parameters-log.json')
        save_parameters_json(filename, self._cutoff, self._activation,
                             self._hiddenlayers, self._label, self._Gs,
                             self._fingerprint_range, ravel, vector)

        self._weights = costfxn._weights
        self._scalings = costfxn._scalings

##############################################################################
##############################################################################
##############################################################################


class FeedForward:

    """
    Class that implements a basic feed-forward neural network.
    Task:  Constructs the feed-forward neural
            network and calculates the output (energy) and also the direvative
            of the energy with respect to coordinates (forces).
    Inputs:
        hiddenlayers: dictionary of chemical element symbols and architectures
                    of their corresponding hidden layers of the conventional
                    neural network. Note that for each atom, number of nodes in
                    the input layer is always equal to the number of symmetry
                    functions (G), and the number of nodes in the output
                    layer is always one. For example,
                    hiddenlayers = {"O":(3,5), "Au":(5,6)} means that for "O"
                    we have two hidden layers, the first one with three nodes
                    and the second one having five nodes.
       weights: dictionary of symbols and dictionaries of arrays of weights;
                   each symbol and related dictionary corresponds to one
                   element and one conventional neural network.  The weights
                   dictionary for the above example is
                   weights = {"O": {1: np.array(149,3), 2: np.array(4,5),
                   3: np.array(6,1)}, "Au": {1: np.array(148,5),
                   2: np.array(6,6), 3: np.array(7,1)}. The arrays are set up
                   to connect node i in the previous layer with node j in the
                   current layer with indices w[i,j]. There are n+1 rows
                   (i values), with the last row being the bias, and m columns
                   (j values). If weights is not given, the arrays will be
                   randomly generated from values between -0.5 and 0.5.
        scalings: dictionary of parameters for slope and intercept for each
                    element. They are used in order to remove the final value
                    from the range that the activation function would
                    otherwise be locked in. For example,
                      scalings={"Pd":{"slope":2, "intercept":1},
                                "O":{"slope":3, "intercept":4}}
        activation: string to assign the type of activation funtion. "linear"
                    refers to linear function, "tanh" refers to tanh function,
                    and "sigmoid" refers to sigmoid function.


    **NOTE: The dimensions of the weight matrix should be consistent with
            hiddenlayers.**
    """

    def __init__(self, hiddenlayers, weights, scalings, activation):
        self._hiddenlayers = hiddenlayers
        self._weights = weights
        self._scalings = scalings
        self._activation = activation
        elements = sorted(self._hiddenlayers.keys())
        self._hiddensizes = {}
        for element in elements:
            structure = self._hiddenlayers[element]
            if isinstance(structure, str):
                structure = structure.split('-')
            elif isinstance(structure, int):
                structure = [structure]
            else:
                structure = list(structure)
            hiddensizes = [int(part) for part in structure]
            self._hiddensizes[element] = hiddensizes

    #########################################################################

    def get_output(self, indexfp, symbol):
        """Given fingerprints of the indexed atom and it's symbol, outputs of
        the neural network are calculated."""

        hiddensizes = self._hiddensizes[symbol]
        weight = self._weights[symbol]
        o = OrderedDict()  # node values
        layer = 1  # input layer
        if fmodules:
            o[1] = fmodules.process_o(o_input=indexfp,
                                      weight_no_row=np.shape(weight[1])[0],
                                      weight_no_col=np.shape(weight[1])[1],
                                      weight=weight[1],
                                      activation_function=self._activation)
        else:
            net = OrderedDict()  # excitation
            ohat = OrderedDict()
            ohat[0] = np.matrix(indexfp + [1.0])
            net[1] = ohat[0] * weight[1]
            if self._activation == 'linear':
                o[1] = net[1]  # linear activation function
            elif self._activation == 'tanh':
                o[1] = np.tanh(net[1])  # tanh activation function
            elif self._activation == 'sigmoid':  # sigmoid activation function
                o[1] = 1. / (1. + np.exp(-net[1]))
            ohat[1] = np.hstack((o[1], np.matrix([[1.0]])))
        for hiddensize in hiddensizes[1:]:
            layer += 1
            if fmodules:
                o[layer] = fmodules.process_o(
                    o_input=o[
                        layer - 1],
                    weight_no_row=np.shape(
                        weight[layer])[0],
                    weight_no_col=np.shape(
                        weight[layer])[1],
                    weight=weight[layer],
                    activation_function=self._activation)
            else:
                net[layer] = ohat[layer - 1] * weight[layer]
                if self._activation == 'linear':
                    o[layer] = net[layer]  # linear activation
                elif self._activation == 'tanh':
                    o[layer] = np.tanh(net[layer])  # tanh activation
                elif self._activation == 'sigmoid':
                    # sigmoid activation
                    o[layer] = 1. / (1. + np.exp(-net[layer]))
                ohat[layer] = np.hstack((o[layer], np.matrix([[1.0]])))
        layer += 1  # output layer
        if fmodules:
            o[layer] = fmodules.process_o(
                o_input=o[
                    layer - 1],
                weight_no_row=np.shape(
                    weight[layer])[0],
                weight_no_col=np.shape(
                    weight[layer])[1],
                weight=weight[layer],
                activation_function=self._activation)
        else:
            net[layer] = float(ohat[layer - 1] * weight[layer])
            if self._activation == 'linear':
                o[layer] = net[layer]  # linear activation function
            elif self._activation == 'tanh':
                o[layer] = np.tanh(net[layer])  # tanh activation function
            elif self._activation == 'sigmoid':
                # sigmoid activation function
                o[layer] = 1. / (1. + np.exp(-net[layer]))
        E = self._scalings[symbol]['slope'] * float(o[layer]) + \
            self._scalings[symbol]['intercept']
        return o, E

    #########################################################################

    def calculate_images_outputs(self, images, sfp):
        """Given the fingerprint of all atoms in all images, the outputs of
        the neural networks are calculated."""

        outputs = OrderedDict()
        NN_energies = OrderedDict()
        for image_no in images.keys():
            NN_energy = 0.
            o = OrderedDict()
            for atom in images[image_no]:
                index = atom.index
                symbol = atom.symbol
                indexfp = sfp.get_fp(image_no, index)
                o[index], ee = self.get_output(indexfp, symbol)
                NN_energy += ee
            outputs[image_no] = o
            NN_energies[image_no] = NN_energy

        return outputs, NN_energies

##############################################################################
##############################################################################
##############################################################################


class CalculateFingerprints:

    """
    Class that calculates fingerprints.
    Inputs:
        cutoff: float (default 6.5 Angstroms)
                  Radius above which neighbor interactions are ignored.
        Gs:  dictionary of symbols and lists of dictionaries for making
                    symmetry functions. Either auto-genetrated, or given
               in the following form:
                    Gs = {"O": [..., ...], "Au": [..., ...]} where
                    ... = {"type":"G2", "element":"O", "eta":0.0009} or
                    ... = {"type":"G4", "elements":["O", "Au"], "eta":0.0001,
                           "gamma":0.1, "zeta":1.0}
        atoms: ASE atoms object
                   The initial atoms object on which the fingerprints will be
                   generated.
    """

    def __init__(self, cutoff, Gs, atoms):
        self._cutoff = cutoff
        self._Gs = Gs
        self._atoms = atoms

        self.update()

    #########################################################################

    def update(self):
        """Update the neighborlist for making fingerprint. Use if atoms
        position has changed."""
        self._nl = NeighborList(cutoffs=([self._cutoff / 2.] *
                                         len(self._atoms)),
                                self_interaction=False,
                                bothways=True,
                                skin=0.)
        self._nl.update(self._atoms)

    #########################################################################

    def index_fingerprint(self, symbol, index):
        """Returns the fingerprint of symmetry function values for atom
        specified by its index. Will automatically update if atom positions
        has changed; you don't need to call update() unless you are
        specifying a new atoms object."""

        n_indices, n_offsets = self._nl.get_neighbors(index)
        symbols = []
        Rs = []
        for n_index, n_offset in zip(n_indices, n_offsets):
            symbols.append(self._atoms[n_index].symbol)
            Rs.append(self._atoms.positions[n_index] +
                      np.dot(n_offset, self._atoms.get_cell()))
        fingerprint = []
        for G in self._Gs[symbol]:
            ridge = self._processG(G, index, symbols, Rs)
            fingerprint.append(ridge)

        return fingerprint

    #########################################################################

    def get_der_fingerprint(self, symbol, index, mm, ii):
        """Returns the derivative of the fingerprint of symmetry function
        values for atom specified by its index with respect to coordinate
        x_{ii} of atom index mm."""

        Rindex = self._atoms.positions[index]
        n_indices, n_offsets = self._nl.get_neighbors(index)
        symbols = []
        Rs = []
        for n_index, n_offset in zip(n_indices, n_offsets):
            symbols.append(self._atoms[n_index].symbol)
            Rs.append(self._atoms.positions[n_index] +
                      np.dot(n_offset, self._atoms.get_cell()))
        der_fingerprint = []
        for G in self._Gs[symbol]:
            ridge = self._process_der_G(G, n_indices, symbols,
                                        Rs, index, Rindex, mm, ii)
            der_fingerprint.append(ridge)
        return der_fingerprint

    #########################################################################

    def _processG(self, G, index, symbols, Rs):
        """Returns the value of G for atom at index. symbols and Rs are
        lists of neighbors' symbols and Cartesian positions, respectively.
        """
        home = self._atoms[index].position

        if G['type'] == 'G2':
            return calculate_G2(symbols, Rs, G['element'], G['eta'],
                                self._cutoff, home)
        elif G['type'] == 'G4':
            return calculate_G4(symbols, Rs, G['elements'], G['gamma'],
                                G['zeta'], G['eta'], self._cutoff, home)
        else:
            raise NotImplementedError('Unknown G type: %s' % G['type'])

    #########################################################################

    def _process_der_G(self, G, n_indices, symbols, Rs, aa, Ra, mm, ii):
        """Returns the value of the derivative of G for atom at index aa and
        position Ra with respect to coordinate x_{ii} of atom index mm.
        Symbols and Rs are lists of neighbors' symbols and Cartesian positions,
        respectively.
        """

        if G['type'] == 'G2':
            return calculate_der_G2(n_indices, symbols, Rs, G['element'],
                                    G['eta'], self._cutoff, aa, Ra, mm, ii)
        elif G['type'] == 'G4':
            return calculate_der_G4(
                n_indices,
                symbols,
                Rs,
                G['elements'],
                G['gamma'],
                G['zeta'],
                G['eta'],
                self._cutoff,
                aa,
                Ra,
                mm,
                ii)
        else:
            raise NotImplementedError('Unknown G type: %s' % G['type'])


##############################################################################
##############################################################################
##############################################################################


class CostFxnandDer:

    """Cost function and it's derivative based on rms difference in energy, to
    be optimized in setting parameters. Initialize with list of images (atoms),
    a ravel function to decode/encode the vector into parameters, and
    convergence criteria (eV/atom).

        cutoff: cutoff radius, in Angstroms, around each atom

        Gs: fingerprint generator function descriptions

        hiddenlayers: structure of hidden layers, e.g., (5, 5)

        ravel: decoder to convert vector to biases / weights and vice versa

        images: list of ASE atoms objects (the training set)

        activation: the type of activation function

        label: str, name used for all files.

        weights: dictionary of element and corresponding nn_structure with
                weights

        converge: threshold rmse/atom at which simulation is converged

        sfp: An object of the class SavedFingerprint, which contains all
            fingerprints

        fp_label: string, name used for saving fingerprint in JSON format.

        fingerprint_range: range of fingerprints of each chemical species

        save_fingerprint_derivatives: boolean, if "True", fingerprint derivatives
        will be saved during force training

    """
    #########################################################################

    def __init__(self, cutoff, Gs, hiddenlayers, ravel, images,
                 activation, label, converge, sfp, fp_label,
                 fingerprint_range):
        self._cutoff = cutoff
        self._Gs = Gs
        self._hiddenlayers = hiddenlayers
        self._ravel = ravel
        self._images = images
        self._activation = activation
        self._label = label
        self._converge = converge
        self._sfp = sfp
        self._fp_label = fp_label
        self._fingerprint_range = fingerprint_range
        self._steps = 0
        self._elements = sorted(self._hiddenlayers.keys())
        filename = make_filename(label, 'train-log.txt')
        self._trainlog = paropen(filename, 'a')
        self._mp = MultiProcess(images=self._images, no_procs=mp.cpu_count())

    #########################################################################

    def f(self, vector):
        """function to calculate the cost function"""

        self._weights, self._scalings = self._ravel.to_dicts(vector)

        ff = FeedForward(self._hiddenlayers, self._weights, self._scalings,
                         self._activation)
        target_args = (ff, self._sfp)

        square_error, self._der_weights_square_error, self._der_scalings_square_error = \
            self._mp.FeedForwardAndBackpropagation_MP(
                target=cost_fxn_derivatives_wrapper,
                sub_args=target_args, weights=self._weights)

        square_error_per_image = square_error / len(self._images)
        rmse_per_atom = np.sqrt(
            square_error_per_image) / len(self._images[0].positions)

        if self._steps == 0:
            head = ('%7s ' ' %25s' ' %15s' '%15s\n')
            self._trainlog.write(
                head %
                ('Step',
                 'Time',
                 'Error, eV^2',
                 'RMSE, meV/atom'))
            self._trainlog.write(head %
                                 ('=' * 6, '=' * 24, '=' * 14, '=' * 14))

        self._steps += 1

        line = ('%7s:' ' %25s' ' %15.2f' '%15.2f\n')
        self._trainlog.write(line % (self._steps - 1, time.ctime(),
                                     square_error, rmse_per_atom * 1000.))
        if self._steps % 1000 == 0:
            self._trainlog.write('Saving checkpoint data.\n')
            filename = make_filename(
                self._label,
                'parameters-checkpoint-%i.json' %
                self._steps)
            save_parameters_json(filename, self._cutoff, self._activation,
                                 self._hiddenlayers, self._label, self._Gs,
                                 self._fingerprint_range, self._ravel, vector)

        if rmse_per_atom < self._converge:
            self._trainlog.write('Convergence!\n')
            self._weights, self._scalings = self._ravel.to_dicts(vector)

            raise ConvergedException

        return square_error

    #########################################################################

    def fprime(self, vector):
        """function to calculate derivative of the cost function"""

        if self._steps == 0:

            self._weights, self._scalings = self._ravel.to_dicts(vector)
            ff = FeedForward(self._hiddenlayers, self._weights, self._scalings,
                             self._activation)
            target_args = (ff, self._sfp)

            square_error, self._der_weights_square_error, self._der_scalings_square_error = \
                self._mp.FeedForwardAndBackpropagation_MP(
                    target=cost_fxn_derivatives_wrapper,
                    sub_args=target_args, weights=self._weights)

        ravel = ParamRavel(self._hiddenlayers, self._der_weights_square_error,
                           self._der_scalings_square_error)
        der_vector = ravel.to_vector(
            self._der_weights_square_error,
            self._der_scalings_square_error)

        return der_vector

##############################################################################
##############################################################################
##############################################################################


class ConvergedException(Exception):

    """Kludge to decide when scipy's optimizers are complete."""
    pass

##############################################################################
##############################################################################
##############################################################################


def calculate_error_per_image_atom(images, NN_energies):
    """
    function to calculate root mean square error (rmse).
    Task:  For given images and neural network energies the rmse per atom per
    image is calculated.

    """
    no_atoms = len(images[0].positions)
    act_energies = []
    _NN_energies = []
    for image_no in images.keys():
        act_energies.append(images[image_no].get_potential_energy(
            apply_constraint=False))
        _NN_energies.append(NN_energies[image_no])
    if fmodules:
        rmse_per_atom, square_error = fmodules.calculate_error(
            no_atom=no_atoms, nn_values=np.array(_NN_energies),
            real_values=np.array(act_energies), no_values=len(images))
    else:
        square_error = 0.
        for image_no in images.keys():
            NN_energy = float(_NN_energies[image_no])
            act_energy = images[image_no].get_potential_energy(
                apply_constraint=False)
            square_error += (NN_energy - act_energy) ** 2.
        square_error_per_image = square_error / len(images)
        rmse_per_atom = np.sqrt(square_error_per_image) / no_atoms

    return rmse_per_atom, square_error

##############################################################################
##############################################################################
##############################################################################


class SavedFingerprint:

    """Memory class to not recalculate fingerprints if not necessary.
    This could cause runaway memory usage; use with caution.
    """

    ##########################################################################

    def __init__(
            self,
            cutoff,
            elements,
            Gs,
            images,
            label,
            fp_label,
            train_forces,
            save_fingerprint_derivatives):
        self._cutoff = cutoff
        self._Gs = Gs
        self._train_forces = train_forces
        self._save_fingerprint_derivatives = save_fingerprint_derivatives
        self._fp_data = {}
        self._der_fp_data = {}

        try:
            filename = make_filename(fp_label, 'fingerprints.json')
            fp = paropen(filename, 'rb')
            data = json.load(fp)
            for key1 in data.keys():
                for key2 in data[key1]:
                    fp_value = data[key1][key2]
                    self._fp_data[(int(key1), int(key2))] = \
                        [float(value) for value in fp_value]
        except:
            fp = None

        prev_image_nos = []
        for key in self._fp_data.keys():
            if key[0] not in prev_image_nos:
                prev_image_nos.append(key[0])

        new_images = {}
        for i in range(len(prev_image_nos), len(images)):
            new_images[i] = images[i]

        if len(new_images) != 0:
            # maximum number of images per queue to be not overloaded
            max_no_images_per_queue = int(
                5 * 10 / len(new_images[new_images.keys()[0]]))
            _mp = MultiProcess(images=new_images, no_procs=mp.cpu_count())
            new_data = _mp.CalculateFingerprints_MP(
                target=fingerprint_calculation_wrapper, sub_args=(cutoff, Gs),
                max_no_images_per_queue=max_no_images_per_queue)
            for key in new_data.keys():
                fp_value = new_data[key]
                self._fp_data[key] = [float(value) for value in fp_value]
        dict_data = OrderedDict()
        keys1 = []
        for key in self._fp_data.keys():
            if key[0] not in keys1:
                keys1.append(key[0])
        for key in keys1:
            dict_data[key] = {}
        for key in self._fp_data.keys():
            dict_data[key[0]][key[1]] = self._fp_data[key]
        filename = make_filename(label, 'fingerprints.json')
        save_fingerprints_json(filename, dict_data)

        fingerprint_values = {}
        for element in elements:
            fingerprint_values[element] = {}
            for i in range(len(Gs[element])):
                fingerprint_values[element][i] = []

        for image_no in images.keys():
            for atom in images[image_no]:
                for i in range(len(Gs[atom.symbol])):
                    fingerprint_values[atom.symbol][i].append(
                        dict_data[image_no][atom.index][i])

        fingerprint_range = OrderedDict()
        for element in elements:
            fingerprint_range[element] = []
            for i in range(len(Gs[element])):
                fingerprint_range[element].append(
                    [min(fingerprint_values[element][i]),
                     max(fingerprint_values[element][i])])

        self._fingerprint_range = fingerprint_range

        if train_forces and save_fingerprint_derivatives:

            try:
                filename = make_filename(
                    fp_label,
                    'fingerprint_derivatives.json')
                fp = paropen(filename, 'rb')
                data = json.load(fp)
                for key1 in data.keys():
                    for key2 in data[key1]:
                        key3 = json.loads(key2)
                        fp_value = data[key1][key2]
                        self._der_fp_data[(int(key1), (int(key3[0]), int(key3[1]), int(key3[2])))] = \
                            [float(value) for value in fp_value]
            except:
                fp = None

            prev_image_nos = []
            for key in self._der_fp_data.keys():
                if key[0] not in prev_image_nos:
                    prev_image_nos.append(key[0])

            new_images = {}
            for i in range(len(prev_image_nos), len(images)):
                new_images[i] = images[i]

            if len(new_images) != 0:
                for image_no in new_images.keys():
                    image = new_images[image_no]
                    cf = CalculateFingerprints(cutoff, Gs, image)
                    for self_atom in image:
                        self_index = self_atom.index
                        for atom in image:
                            index = atom.index
                            symbol = atom.symbol
                            for i in range(3):
                                der_indexfp = cf.get_der_fingerprint(
                                    symbol,
                                    index,
                                    self_index,
                                    i)
                                self._der_fp_data[
                                    (image_no, (index, self_index, i))] = der_indexfp

            dict_data = OrderedDict()
            keys1 = []
            for key in self._der_fp_data.keys():
                if key[0] not in keys1:
                    keys1.append(key[0])
            for key in keys1:
                dict_data[key] = {}
            for key in self._der_fp_data.keys():
                dict_data[key[0]][key[1]] = self._der_fp_data[key]
            filename = make_filename(label, 'fingerprint_derivatives.json')
            save_der_fingerprints_json(filename, dict_data)

    #########################################################################

    def get_fp(self, image_no, atom_no):
        """Returns data if it exists; None if not."""
        if (image_no, atom_no) in self._fp_data.keys():
            return self._fp_data[(image_no, atom_no)]

    #########################################################################

    def get_der_fp(self, image_no, xxx_todo_changeme):
        """Returns data if it exists; None if not."""
        (index, self_index, i) = xxx_todo_changeme
        if (image_no, (index, self_index, i)) in self._der_fp_data.keys():
            return self._der_fp_data[(image_no, (index, self_index, i))]

##############################################################################
##############################################################################
##############################################################################


class ParamRavel:

    """Class to ravel and unravel parameter values into a single vector.
    This is used for feeding into the optimizer. Feed in a list of
    dictionaries to initialize. The dictionaries should be two deep.
        hiddenlayers: structure of hiddenlayers, e.g., (5, 5)
        Gs: fingerprint generator function descriptions
        weights, scalings: parameters to ravel and unravel
    """

    ##########################################################################

    def __init__(self, hiddenlayers, weights, scalings):

        count = 0
        self._weightskeys = []
        self._scalingskeys = []
        for key1 in sorted(weights.keys()):  # element
            for key2 in sorted(weights[key1].keys()):  # layer
                value = weights[key1][key2]
                self._weightskeys.append({'key1': key1,
                                          'key2': key2,
                                          'shape': np.array(value).shape,
                                          'size': np.array(value).size})
                count += np.array(weights[key1][key2]).size
        for key1 in sorted(scalings.keys()):  # element
            for key2 in sorted(scalings[key1].keys()):  # slope / intercept
                self._scalingskeys.append({'key1': key1,
                                           'key2': key2})
                count += 1
        self._vector = np.zeros(count)

    #########################################################################

    def to_vector(self, weights, scalings):
        """Puts the weights and scalings embedded dictionaries into a single
        vector and returns it. The bias and weights dictionary need to have
        the identical structure to those it was initialized with."""
        count = 0
        for k in sorted(self._weightskeys):
            lweights = np.array(weights[k['key1']][k['key2']]).ravel()
            self._vector[count:(count + lweights.size)] = lweights
            count += lweights.size
        for k in sorted(self._scalingskeys):
            self._vector[count] = scalings[k['key1']][k['key2']]
            count += 1
        return self._vector

    #########################################################################

    def to_dicts(self, vector):
        """Puts the vector back into weights and scalings dictionaries of the
        form initialized. vector must have same length as the output of
        unravel."""

        assert len(vector) == len(self._vector)
        self._vector = vector
        count = 0
        weights = OrderedDict()
        scalings = OrderedDict()
        for k in sorted(self._weightskeys):
            if k['key1'] not in weights.keys():
                weights[k['key1']] = OrderedDict()
            matrix = vector[count:count + k['size']]
            matrix = matrix.flatten()
            matrix = np.matrix(matrix.reshape(k['shape']))
            weights[k['key1']][k['key2']] = matrix
            count += k['size']
        for k in sorted(self._scalingskeys):
            if k['key1'] not in scalings.keys():
                scalings[k['key1']] = OrderedDict()
            scalings[k['key1']][k['key2']] = vector[count]
            count += 1
        return weights, scalings

##############################################################################
##############################################################################
##############################################################################


class GetForces:

    """
    Class to calculate coordinate derivatives of energy (forces).

    Inputs:
        cutoff: float (default 6.5 Angstroms)
                  Radius above which neighbor interactions are ignored.
        Gs:  dictionary of symbols and lists of dictionaries for making
                    symmetry functions. Either auto-genetrated, or given
               in the following form:
                    Gs = {"O": [..., ...], "Au": [..., ...]} where
                    ... = {"type":"G2", "element":"O", "eta":0.0009} or
                    ... = {"type":"G4", "elements":["O", "Au"], "eta":0.0001,
                           "gamma":0.1, "zeta":1.0}
        atoms: ASE atoms object
                   The initial atoms object on which the fingerprints will be
                   generated.
        hiddenlayers: dictionary of chemical element symbols and architectures
                    of their corresponding hidden layers of the conventional
                    neural network. Note that for each atom, number of nodes in
                    the input layer is always equal to the number of symmetry
                    functions (G), and the number of nodes in the output
                    layer is always one. For example,
                    hiddenlayers = {"O":(3,5), "Au":(5,6)} means that for "O"
                    we have two hidden layers, the first one with three nodes
                    and the second one having five nodes.
       weights: dictionary of symbols and dictionaries of arrays of weights;
               each symbol and related dictionary corresponds to one chemical
                   element and one conventional neural network.  The weights
                   dictionary for the above example is
                   weights = {"O": {1: np.array(149,3), 2: np.array(4,5),
                   3: np.array(6,1)}, "Au": {1: np.array(148,5),
                   2: np.array(6,6), 3: np.array(7,1)}. The arrays are set up
                   to connect node i in the previous layer with node j in the
                   current layer with indices w[i,j]. There are n+1 rows
                   (i values), with the last row being the bias, and m columns
                   (j values). If weights is not given, the arrays will be
                   randomly generated from values between -0.5 and 0.5.
        scalings: dictionary of parameters for slope and intercept for each
                    element. They are used in order to remove the final value
                    from the range that the activation function would
                    otherwise be locked in. For example,
                      scalings={"Pd":{"slope":2, "intercept":1},
                                "O":{"slope":3, "intercept":4}}
        activation: string to assign the type of activation funtion. "linear"
                    refers to linear function, "tanh" refers to tanh function,
                    and "sigmoid" refers to sigmoid function.

    **NOTE: The dimensions of the weight matrix should be consistent with the
            hiddenlayers.**
    """

    ##########################################################################

    def __init__(self, cutoff, Gs, atoms, hiddenlayers, weights, scalings,
                 activation):
        self._cutoff = cutoff
        self._Gs = Gs
        self._atoms = atoms
        self._hiddenlayers = hiddenlayers
        self._weights = weights
        self._scalings = scalings
        self._activation = activation

        if self._atoms:
            self._nl = NeighborList(
                cutoffs=[
                    self._cutoff / 2.] * len(
                    self._atoms),
                self_interaction=False,
                bothways=True,
                skin=0.)

        self._elements = sorted(self._hiddenlayers.keys())
        self._hiddensizes = {}
        for element in self._elements:
            structure = self._hiddenlayers[element]
            if isinstance(structure, str):
                structure = structure.split('-')
            else:
                structure = list(structure)
            self._hiddensizes[element] = [int(part) for part in structure]

        self._cf = CalculateFingerprints(self._cutoff, self._Gs, self._atoms)

        self._outputs, NN_energy = self.calculate_outputs()

    ##########################################################################

    def calculate_force(self, mm, ii):
        """ Calculates the force in ii direction on atom index mm"""

        force = 0.
        for atom in self._atoms:
            symbol = atom.symbol
            index = atom.index
            o = self._outputs[index]
            index_der_fp = self._cf.get_der_fingerprint(symbol, index, mm, ii)
            hiddensizes = self._hiddensizes[symbol]
            weight = self._weights[symbol]
            der_o = {}  # node values
            der_o[0] = index_der_fp
            layer = 0  # input layer
            for hiddensize in hiddensizes[0:]:
                layer += 1
                temp = np.dot(np.matrix(der_o[layer - 1]),
                              np.delete(weight[layer], -1, 0))
                der_o[layer] = []
                for j in range(len(o[layer])):
                    if self._activation == 'linear':  # linear function
                        der_o[layer].append(float(temp[0, j]))
                    elif self._activation == 'sigmoid':  # sigmoid function
                        der_o[layer].append(
                            float(o[layer][j] * (1. - o[layer][j])) *
                            float(temp[0, j]))
                    elif self._activation == 'tanh':  # tanh function
                        der_o[layer].append(
                            float(1. - o[layer][j] * o[layer][j]) *
                            float(temp[0, j]))
            layer += 1  # output layer
            temp = np.dot(
                np.matrix(der_o[layer - 1]), np.delete(weight[layer], -1, 0))
            if self._activation == 'linear':  # linear function
                der_o[layer] = float(temp)
            elif self._activation == 'sigmoid':  # sigmoid function
                der_o[layer] = float(o[layer] * (1. - o[layer]) * temp)
            elif self._activation == 'tanh':  # tanh function
                der_o[layer] = float((1. - o[layer] * o[layer]) * temp)
            force += float(-(self._scalings[symbol]['slope'] * der_o[layer]))

        return force

    #########################################################################

    def calculate_outputs(self):
        NN_energy = 0.
        o = {}
        cf = CalculateFingerprints(self._cutoff, self._Gs, self._atoms)
        ff = FeedForward(self._hiddenlayers, self._weights,
                         self._scalings, self._activation)
        for atom in self._atoms:
            index = atom.index
            symbol = atom.symbol
            indexfp = cf.index_fingerprint(symbol, index)
            o[index], ee = ff.get_output(indexfp, symbol)
            NN_energy += ee
        return o, NN_energy

    #########################################################################

    def update(self):
        """Update the neighborlist for making fingerprint. Use if atoms
        position has changed."""
        if not self._atoms:
            self._nl = NeighborList(cutoffs=([self._cutoff / 2.] *
                                             len(self._atoms)),
                                    self_interaction=False,
                                    bothways=True,
                                    skin=0.)
        self._nl.update(self._atoms)

##############################################################################


def make_weight_matrices(Gs, elements, hiddenlayers, activation):
    """Makes random weight matrices from parameters according to the convention
    used in Behler (J. Chem. Physc.: atom-centered symmetry functions for
    constructing ...).
    """
    if activation == 'linear':
        weight_range = 0.1
    else:
        weight_range = 1.

    W = OrderedDict()
    nn_structure = OrderedDict()
    for element in sorted(elements):
        if isinstance(hiddenlayers[element], int):
            nn_structure[element] = ([len(Gs[element])] +
                                     [hiddenlayers[element]] +
                                     [1])
        else:
            nn_structure[element] = (
                [len(Gs[element])] +
                [layer for layer in hiddenlayers[element]] + [1])
        W[element] = OrderedDict()
        W[element][1] = np.random.random(
            (len(
                Gs[element]) + 1,
                nn_structure[element][1])) * weight_range - weight_range / 2.
        for layer in range(len(list(nn_structure[element])) - 3):
            W[element][
                layer + 2] = np.random.random(
                (nn_structure[element][
                    layer + 1] + 1,
                    nn_structure[element][
                    layer + 2])) * weight_range - weight_range / 2.
        W[element][len(list(nn_structure[element])) - 1] = np.random.random(
            (nn_structure[element][-2] + 1, 1)) \
            * weight_range - weight_range / 2.
        for i in range(len(W[element])):  # biases
            size = W[element][i + 1][-1].size
            for jj in range(size):
                W[element][i + 1][-1][jj] = 0.
    return W

##############################################################################


def make_scalings_matrices(elements, images, activation):
    """Makes initial scaling matrices, such that the range of activation
        is scaled to the range of actual energies.
    """
    no_images = range(len(images))
    max_act_energy = max(images[i].get_potential_energy(apply_constraint=False)
                         for i in no_images)
    min_act_energy = min(images[i].get_potential_energy(apply_constraint=False)
                         for i in no_images)
    scaling = OrderedDict()
    for element in sorted(elements):
        scaling[element] = OrderedDict()
        if activation == 'sigmoid':  # sigmoid activation function
            scaling[element]['intercept'] = (
                min_act_energy) / len(images[0].positions)
            scaling[element]['slope'] = (
                max_act_energy - min_act_energy) / len(images[0].positions)
        elif activation == 'tanh':  # tanh activation function
            scaling[element]['intercept'] = (max_act_energy + min_act_energy) \
                / 2. / len(images[0].positions)
            scaling[element]['slope'] = (max_act_energy - min_act_energy) \
                / 2. / len(images[0].positions)
        elif activation == 'linear':  # linear activation function
            scaling[element]['intercept'] = (max_act_energy + min_act_energy)\
                / 2. / len(images[0].positions)
            scaling[element]['slope'] = (
                10. ** (-10.)) * (max_act_energy - min_act_energy) \
                / 2. / len(images[0].positions)
    return scaling

##############################################################################


def calculate_G2(symbols, Rs, G_element, eta, cutoff, home):
    """Calculate G2 symmetry function. Ideally this will not be used but
    will be a template for how to build the fortran version (and serves as
    a slow backup if the fortran one goes uncompiled)."""

    if fmodules:  # fortran version; faster
        G_number = [atomic_numbers[G_element]]
        numbers = [atomic_numbers[symbol] for symbol in symbols]
        ridge = fmodules.calculate_g2(numbers=numbers, rs=Rs,
                                      g_number=G_number, g_eta=eta,
                                      cutoff=cutoff, home=home)
        return ridge
    else:
        ridge = 0.  # One aspect of a fingerprint :)
        for symbol, R in zip(symbols, Rs):
            if symbol == G_element:
                Rij = np.linalg.norm(R - home)
                ridge += (np.exp(-eta * Rij ** 2.) *
                          cutoff_fxn(Rij, cutoff))
        return ridge

##############################################################################


def calculate_G4(symbols, Rs, G_elements, gamma, zeta, eta, cutoff, home):
    """Calculate G4 symmetry function. Ideally this will not be used but
    will be a template for how to build the fortran version (and serves as
    a slow backup if the fortran one goes uncompiled)."""

    if fmodules:  # fortran version; faster
        G_numbers = sorted([atomic_numbers[el] for el in G_elements])
        numbers = [atomic_numbers[symbol] for symbol in symbols]
        ridge = fmodules.calculate_g4(numbers=numbers, rs=Rs,
                                      g_numbers=G_numbers, g_gamma=gamma,
                                      g_zeta=zeta, g_eta=eta,
                                      cutoff=cutoff, home=home)
        return ridge
    else:
        ridge = 0.
        counts = range(len(symbols))
        for j in counts:
            for k in counts[(j + 1):]:
                els = sorted([symbols[j], symbols[k]])
                if els != G_elements:
                    continue
                Rij_ = Rs[j] - home
                Rij = np.linalg.norm(Rij_)
                Rik_ = Rs[k] - home
                Rik = np.linalg.norm(Rik_)
                Rjk = np.linalg.norm(Rs[j] - Rs[k])
                cos_theta_ijk = np.dot(Rij_, Rik_) / Rij / Rik
                term = (1. + gamma * cos_theta_ijk) ** zeta
                term *= np.exp(-eta * (Rij ** 2. + Rik ** 2. + Rjk ** 2.))
                term *= cutoff_fxn(Rij, cutoff)
                term *= cutoff_fxn(Rik, cutoff)
                term *= cutoff_fxn(Rjk, cutoff)
                ridge += term
        ridge *= 2. ** (1. - zeta)
        return ridge

##############################################################################


def make_symmetry_functions(elements, G):
    """Makes symmetry functions as in Nano Letters function by Artrith.
    Elements is a list of the elements, as in ["C", "O", "H", "Cu"].
    G[0] = {"type":"G2", "element": "O", "eta": 0.0009}
    G[40] = {"type":"G4", "elements": ["O", "Au"], "eta": 0.0001,
             "gamma": 1.0, "zeta": 1.0}
    If G (a list) is fed in, this will add to it and return an expanded
    version. If not, it will create a new one.
    """

    if not G:
        G = OrderedDict()
        for element0 in elements:
            GG = []
            # Radial symmetry functions.
            etas = [0.0009, 0.01, 0.02, 0.035, 0.06, 0.1, 0.2]
            for eta in etas:
                for element in elements:
                    GG.append({'type': 'G2', 'element': element, 'eta': eta})
            # Angular symmetry functions.
            etas = [0.0001, 0.003, 0.008]
            gammas = [+1., -1.]
            zetas = [1., 4.]
            for eta in etas:
                for zeta in zetas:
                    for gamma in gammas:
                        for i1, el1 in enumerate(elements):
                            for el2 in elements[i1:]:
                                els = sorted([el1, el2])
                                GG.append({'type': 'G4',
                                           'elements': els,
                                           'eta': eta,
                                           'gamma': gamma,
                                           'zeta': zeta})
            G[element0] = GG
    return G

##############################################################################


def cutoff_fxn(Rij, Rc):
    """Cosine cutoff function in Parinello-Behler method."""
    if Rij > Rc:
        return 0.
    else:
        return 0.5 * (np.cos(np.pi * Rij / Rc) + 1.)

##############################################################################


def der_cutoff_fxn(Rij, Rc):
    """Derivative of the Cosine cutoff function in Parinello-Behler method."""
    if Rij > Rc:
        return 0.
    else:
        return -0.5 * np.pi / Rc * np.sin(np.pi * Rij / Rc)

##############################################################################


def Kronecker_delta(ii, jj):
    """Kronecker delta function."""
    if ii == jj:
        return 1.
    else:
        return 0.

##############################################################################


def der_position_vector(aa, bb, mm, ii):
    """Returns the derivative of the position vector R_{ab} with respect to
        x_{ii} of atomic index mm."""
    der_position_vector = []
    der_position_vector.append(
        (Kronecker_delta(
            mm,
            aa) -
            Kronecker_delta(
            mm,
            bb)) *
        Kronecker_delta(
            0,
            ii))
    der_position_vector.append(
        (Kronecker_delta(
            mm,
            aa) -
            Kronecker_delta(
            mm,
            bb)) *
        Kronecker_delta(
            1,
            ii))
    der_position_vector.append(
        (Kronecker_delta(
            mm,
            aa) -
            Kronecker_delta(
            mm,
            bb)) *
        Kronecker_delta(
            2,
            ii))
    return der_position_vector

##############################################################################


def der_position(mm, nn, Rm, Rn, ll, ii):
    """Returns the derivative of the norm of position vector R_{mn} with
        respect to x_{ii} of atomic index ll."""
    Rmn = np.linalg.norm(Rm - Rn)
    if ll == mm:
        der_position = (Rm[ii] - Rn[ii]) / Rmn
    elif ll == nn:
        der_position = -(Rm[ii] - Rn[ii]) / Rmn
    else:
        der_position = 0.
    return der_position

##############################################################################


def der_cos_theta(aa, jj, kk, Ra, Rj, Rk, mm, ii):
    """Returns the derivative of Cos(theta_{ajk}) with respect to
        x_{ii} of atomic index mm."""
    Raj_ = Ra - Rj
    Raj = np.linalg.norm(Raj_)
    Rak_ = Ra - Rk
    Rak = np.linalg.norm(Rak_)
    der_cos_theta = 1. / \
        (Raj * Rak) * np.dot(der_position_vector(aa, jj, mm, ii), Rak_)
    der_cos_theta += +1. / \
        (Raj * Rak) * np.dot(Raj_, der_position_vector(aa, kk, mm, ii))
    der_cos_theta += -1. / \
        ((Raj ** 2.) * Rak) * np.dot(Raj_, Rak_) * \
        der_position(aa, jj, Ra, Rj, mm, ii)
    der_cos_theta += -1. / \
        (Raj * (Rak ** 2.)) * np.dot(Raj_, Rak_) * \
        der_position(aa, kk, Ra, Rk, mm, ii)
    return der_cos_theta

##############################################################################


def calculate_der_G2(n_indices, symbols, Rs, G_element, eta, cutoff, aa, Ra,
                     mm, ii):
    """Calculate coordinate derivative of G2 symmetry function for atom at
    index aa and position Ra with respect to coordinate x_{ii} of atom index
    mm."""

    if fmodules:  # fortran version; faster
        G_number = [atomic_numbers[G_element]]
        numbers = [atomic_numbers[symbol] for symbol in symbols]
        list_n_indices = [n_indices[i] for i in range(len(n_indices))]
        ridge = fmodules.calculate_der_g2(n_indices=list_n_indices,
                                          numbers=numbers, rs=Rs,
                                          g_number=G_number, g_eta=eta,
                                          cutoff=cutoff, aa=aa, home=Ra, mm=mm,
                                          ii=ii)
    else:
        ridge = 0.  # One aspect of a fingerprint :)
        for symbol, Rj, n_index in zip(symbols, Rs, n_indices):
            if symbol == G_element:
                Raj = np.linalg.norm(Ra - Rj)
                term1 = (-2. * eta * Raj * cutoff_fxn(Raj, cutoff) +
                         der_cutoff_fxn(Raj, cutoff))
                term2 = der_position(aa, n_index, Ra, Rj, mm, ii)
                ridge += np.exp(- eta * Raj ** 2.) * term1 * term2
    return ridge

##############################################################################


def calculate_der_G4(n_indices, symbols, Rs, G_elements, gamma, zeta, eta,
                     cutoff, aa, Ra, mm, ii):
    """Calculate coordinate derivative of G4 symmetry function for atom at
    index aa and position Ra with respect to coordinate x_{ii} of atom index
    mm."""

    if fmodules:  # fortran version; faster
        G_numbers = sorted([atomic_numbers[el] for el in G_elements])
        numbers = [atomic_numbers[symbol] for symbol in symbols]
        list_n_indices = [n_indices[i] for i in range(len(n_indices))]
        ridge = fmodules.calculate_der_g4(n_indices=list_n_indices,
                                          numbers=numbers, rs=Rs,
                                          g_numbers=G_numbers, g_gamma=gamma,
                                          g_zeta=zeta, g_eta=eta,
                                          cutoff=cutoff, aa=aa, home=Ra, mm=mm,
                                          ii=ii)
    else:
        ridge = 0.
        counts = range(len(symbols))
        for j in counts:
            for k in counts[(j + 1):]:
                els = sorted([symbols[j], symbols[k]])
                if els != G_elements:
                    continue
                Rj = Rs[j]
                Rk = Rs[k]
                Raj_ = Rs[j] - Ra
                Raj = np.linalg.norm(Raj_)
                Rak_ = Rs[k] - Ra
                Rak = np.linalg.norm(Rak_)
                Rjk_ = Rs[j] - Rs[k]
                Rjk = np.linalg.norm(Rjk_)
                cos_theta_ajk = np.dot(Raj_, Rak_) / Raj / Rak
                c1 = (1. + gamma * cos_theta_ajk)
                c2 = cutoff_fxn(Raj, cutoff)
                c3 = cutoff_fxn(Rak, cutoff)
                c4 = cutoff_fxn(Rjk, cutoff)
                if zeta == 1:
                    term1 = np.exp(-eta * (Raj ** 2. + Rak ** 2. + Rjk ** 2.))
                else:
                    term1 = c1 ** (zeta - 1.) * \
                        np.exp(-eta * (Raj ** 2. + Rak ** 2. + Rjk ** 2.))
                term2 = c2 * c3 * c4
                term3 = der_cos_theta(aa, n_indices[j], n_indices[k], Ra, Rj,
                                      Rk, mm, ii)
                term4 = gamma * zeta * term3
                term5 = der_position(aa, n_indices[j], Ra, Rj, mm, ii)
                term4 += -2. * c1 * eta * Raj * term5
                term6 = der_position(aa, n_indices[k], Ra, Rk, mm, ii)
                term4 += -2. * c1 * eta * Rak * term6
                term7 = der_position(n_indices[j], n_indices[k], Rj, Rk, mm,
                                     ii)
                term4 += -2. * c1 * eta * Rjk * term7
                term2 = term2 * term4
                term8 = c1 * der_cutoff_fxn(Raj, cutoff) * c3 * c4 * term5
                term9 = c1 * c2 * der_cutoff_fxn(Rak, cutoff) * c4 * term6
                term10 = c1 * c2 * c3 * der_cutoff_fxn(Rjk, cutoff) * term7
                term11 = term2 + term8 + term9 + term10
                term = term1 * term11
                ridge += term
        ridge *= 2. ** (1. - zeta)

    return ridge

##############################################################################


def save_fingerprints_json(filename, fingerprints):
    """Save checkpoints, etc. This is a rough file format that we should
    improve upon when we get a real data format!"""

    new_dict = {}
    for key1 in fingerprints.keys():
        new_dict[str(key1)] = {}
        for key2 in fingerprints[key1].keys():
            fp_value = fingerprints[key1][key2]
            new_dict[str(key1)][str(key2)] = [str(value) for value in fp_value]

    with paropen(filename, 'wb') as outfile:
        json.dump(new_dict, outfile)

##############################################################################


def save_der_fingerprints_json(filename, der_fingerprints):
    """Save checkpoints, etc. This is a rough file format that we should
    improve upon when we get a real data format!"""

    new_dict = {}
    for key1 in der_fingerprints.keys():
        new_dict[str(key1)] = {}
        for key2 in der_fingerprints[key1].keys():
            fp_value = der_fingerprints[key1][key2]
            new_dict[str(key1)][str([key2[0], key2[1], key2[2]])] = [
                str(value) for value in fp_value]

    with paropen(filename, 'wb') as outfile:
        json.dump(new_dict, outfile)

##############################################################################


def save_parameters_json(filename, cutoff, activation, hiddenlayers, label, Gs,
                         fingerprint_range, ravel, vector):
    """Save checkpoints, etc. This is a rough file format that we should
    improve upon when we get a real data format!"""

    parameters = OrderedDict()
    parameters['cutoff'] = cutoff
    parameters['activation'] = activation
    parameters['hiddenlayers'] = hiddenlayers
    parameters['label'] = label
    parameters['Gs'] = Gs
    parameters['fingerprint_range'] = fingerprint_range

    weight, scaling = ravel.to_dicts(vector)

    new_weights = OrderedDict()
    for key1 in sorted(weight.keys()):
        new_weights[key1] = OrderedDict()
        dictionary = weight[key1]
        for key2 in sorted(dictionary.keys()):
            shape = np.shape(dictionary[key2])
            new_weights[key1][str(key2)] = []
            for i in range(shape[0]):
                new_weights[key1][str(key2)].append([])
                for j in range(shape[1]):
                    value = float(dictionary[key2][i, j])
                    new_weights[key1][str(key2)][i].append(value)

    parameters['weights'] = new_weights
    parameters['scalings'] = scaling

    base_filename = os.path.splitext(filename)[0]
    export_filename = os.path.join(base_filename + '.json')

    with paropen(export_filename, 'wb') as outfile:
        json.dump(parameters, outfile)

##############################################################################


def read_parameters_json(json_file):
    """Reads parameters from JSON file."""

    parameters = json.load(json_file)
    parameters_weights = parameters['weights']

    new_weights = OrderedDict()
    for key1 in sorted(parameters_weights.keys()):
        new_weights[key1] = OrderedDict()
        dictionary = parameters_weights[key1]
        for key2 in sorted(dictionary.keys()):
            shape = np.shape(dictionary[key2])
            new_weights[key1][int(key2)] = []
            for i in range(shape[0]):
                new_weights[key1][int(key2)].append([])
                for j in range(shape[1]):
                    value = float(dictionary[key2][i][j])
                    new_weights[key1][int(key2)][i].append(value)

    parameters_scalings = parameters['scalings']

    new_scalings = OrderedDict()
    for element in sorted(parameters_scalings.keys()):
        new_scalings[element] = OrderedDict()
        new_scalings[element]['intercept'] = \
            parameters_scalings[element]['intercept']
        new_scalings[element]['slope'] = parameters_scalings[element]['slope']

    weights = new_weights
    scalings = new_scalings
    cutoff = parameters['cutoff']
    activation = parameters['activation']
    hiddenlayers = parameters['hiddenlayers']
    label = parameters['label']
    Gs = parameters['Gs']
    fingerprint_range = parameters['fingerprint_range']

    return cutoff, activation, hiddenlayers, label, Gs, \
        fingerprint_range, weights, scalings

##############################################################################


def make_filename(label, base_filename):
    """Creates a filename from the label and the base_filename which should be
    a string"""

    if not label:
        filename = base_filename
    else:
        filename = os.path.join(label + '-' + base_filename)

    return filename

##############################################################################


def make_hiddenlayer_structure(elements):
    """Creates a dictionary of elements as the keys and corresponding
    structures"""

    hiddenlayers = OrderedDict()
    for element in elements:
        hiddenlayers[element] = (5, 5)
    return hiddenlayers

##############################################################################
##############################################################################
##############################################################################


def make_list_sub_images(images, max_no_queues=None, no_procs=None):
    """ Generates list of sub images.  The total length of the created list is
     no_procs if only no_procs is given, however, if max_no_queue is given then
     the length of sub-images is max_no_queues"""

    if no_procs:

        min_image_no = min(images.keys())
        len_sub_images = int(len(images) / no_procs)
        list_sub_images = []
        for i in range(no_procs - 1):
            sub_images = {}
            for j in range(len_sub_images):
                sub_images[
                    min_image_no +
                    i *
                    len_sub_images +
                    j] = images[
                    min_image_no +
                    i *
                    len_sub_images +
                    j]
            list_sub_images.append(sub_images)
        last_sub_image = {}

        for i in range(len_sub_images * (no_procs - 1), len(images)):
            last_sub_image[min_image_no + i] = images[min_image_no + i]
        list_sub_images.append(last_sub_image)
        return list_sub_images

    elif max_no_queues:

        min_image_no = min(images.keys())
        list_sub_images = []
        i = -1
        while i < len(images) / max_no_queues - 1:
            i += 1
            sub_images = {}
            for j in range(max_no_queues):
                sub_images[
                    min_image_no +
                    i *
                    max_no_queues +
                    j] = images[
                    min_image_no +
                    i *
                    max_no_queues +
                    j]
            list_sub_images.append(sub_images)
        if (i + 1) * max_no_queues < len(images):
            last_sub_image = {}
            for k in range((i + 1) * max_no_queues, len(images)):
                last_sub_image[min_image_no + k] = images[min_image_no + k]
            list_sub_images.append(last_sub_image)
        return list_sub_images

##############################################################################
##############################################################################
##############################################################################


class MultiProcess:

    """Class to do parallel processing, using multiprocessing package which
    works on Python versions 2.6 and above.
    Inputs:
            images: dictionary of image objects
            no_procs: number of processors
            """

    def __init__(self, images, no_procs):

        self._images = images
        self._no_procs = no_procs
        self._queues = {}
        for x in range(no_procs):
            self._queues[x] = mp.Queue()

    ##########################################################################

    def CalculateFingerprints_MP(self, target, sub_args,
                                 max_no_images_per_queue):
        """Fingerprints are calculated in parallel"""

        list_sub_images = make_list_sub_images(
            images=self._images, max_no_queues=max_no_images_per_queue)
        data = {}
        y = 0
        while (y + 1) * self._no_procs <= len(list_sub_images) - 1:
            y = y + 1
            args = {}
            for x in range(self._no_procs):
                index = (y - 1) * self._no_procs + x
                args[x] = (list_sub_images[index],) + sub_args + \
                    (self._queues[x],)

            processes = [mp.Process(target=target, args=args[x])
                         for x in range(self._no_procs)]

            for x in range(self._no_procs):
                processes[x].start()

            for x in range(self._no_procs):
                processes[x].join()

            sub_data = []

            for x in range(self._no_procs):
                result = self._queues[x].get()
                sub_data.append(result)

#       Construct all image fingerprints from subprocesses
            for i in range(len(sub_data)):
                for key in sub_data[i].keys():
                    data[key] = sub_data[i][key]

        for x in range(y * self._no_procs, len(list_sub_images)):
            self._queues[x] = mp.Queue()

        args = {}
        for x in range(y * self._no_procs, len(list_sub_images)):
            args[x] = (list_sub_images[x],) + sub_args + (self._queues[x],)

        processes = [mp.Process(target=target, args=args[x])
                     for x in range(y * self._no_procs, len(list_sub_images))]

        for x in range(len(processes)):
            processes[x].start()

        for x in range(len(processes)):
            processes[x].join()

        sub_data = []
        for x in range(y * self._no_procs, len(list_sub_images)):
            result = self._queues[x].get()
            sub_data.append(result)

#       Construct all image fingerprints from subprocesses
        for i in range(len(sub_data)):
            for key in sub_data[i].keys():
                data[key] = sub_data[i][key]

        return data

    ##########################################################################

    def FeedForwardAndBackpropagation_MP(self, target, sub_args, weights):
        """Derivatives of the cost function with respect to parameters are
        calculated in parallel"""

        list_sub_images = make_list_sub_images(images=self._images,
                                               no_procs=self._no_procs)
        args = {}
        for x in range(self._no_procs):
            args[x] = (list_sub_images[x],) + sub_args + (self._queues[x],)

        square_error = 0.

        der_weights_square_error = OrderedDict()
        for key1 in sorted(weights.keys()):
            der_weights_square_error[key1] = {}
            for key2 in weights[key1].keys():
                shape = np.shape(weights[key1][key2])
                der_weights_square_error[key1][key2] = np.zeros(shape=shape)

        der_scalings_square_error = OrderedDict()
        for key1 in sorted(weights.keys()):
            der_scalings_square_error[key1] = {}
            der_scalings_square_error[key1]['intercept'] = 0.
            der_scalings_square_error[key1]['slope'] = 0.

        processes = [mp.Process(target=target, args=args[x])
                     for x in range(self._no_procs)]

        for x in range(self._no_procs):
            processes[x].start()

        for x in range(self._no_procs):
            processes[x].join()

        sub_square_error = []
        sub_der_weights_square_error = []
        sub_der_scalings_square_error = []

#       Construct total square_error and derivative with respect to parameters
#       from subprocesses
        for x in range(self._no_procs):
            result = self._queues[x].get()
            sub_square_error.append(result[0])
            sub_der_weights_square_error.append(result[1])
            sub_der_scalings_square_error.append(result[2])

        for i in range(len(sub_square_error)):
            square_error += sub_square_error[i]

        for key1 in weights.keys():
            for key2 in weights[key1].keys():
                shape = np.shape(weights[key1][key2])
                for m in range(shape[0]):
                    for n in range(shape[1]):
                        for x in range(self._no_procs):
                            der_weights_square_error[key1][key2][m][
                                n] += sub_der_weights_square_error[x][key1][key2][m][n]

        for key1 in weights.keys():
            for x in range(self._no_procs):
                der_scalings_square_error[key1]['intercept'] += \
                    sub_der_scalings_square_error[x][key1]['intercept']
                der_scalings_square_error[key1]['slope'] += \
                    sub_der_scalings_square_error[x][key1]['slope']

        return square_error, der_weights_square_error, der_scalings_square_error


##############################################################################

def fingerprint_calculation_wrapper(images, cutoff, Gs, queues):
    """wrapper function to be used in multiprocessing for calculating
    fingerprints"""

    data = {}
    for image_no in images.keys():
        image = images[image_no]
        cf = CalculateFingerprints(cutoff, Gs, image)
        for atom in image:
            index = atom.index
            symbol = atom.symbol
            indexfp = cf.index_fingerprint(symbol, index)
            data[(image_no, index)] = indexfp

    queues.put(data)

##############################################################################


def der_fingerprint_calculation_wrapper(images, cutoff, Gs, queues):
    """wrapper function to be used in multiprocessing for calculating
    derivatives of fingerprints"""

    data = {}
    for image_no in images.keys():
        image = images[image_no]
        cf = CalculateFingerprints(cutoff, Gs, image)
        for self_atom in image:
            self_index = self_atom.index
            for atom in image:
                index = atom.index
                symbol = atom.symbol
                for i in range(3):
                    index_der_fp = cf.get_der_fingerprint(
                        symbol,
                        index,
                        self_index,
                        i)
                    data[(image_no, (index, self_index, i))] = index_der_fp

    queues.put(data)

##############################################################################


def cost_fxn_derivatives_wrapper(images, ff, sfp, queues):
    """wrapper function to be used in multiprocessing for calculating
    cost function and it's derivative with respect to parameters"""

    elements = sorted(ff._hiddenlayers.keys())
    activation = ff._activation
    weights = ff._weights
    scalings = ff._scalings

    square_error = 0.

    der_weights_square_error = OrderedDict()

    for element in elements:
        der_weights_square_error[element] = OrderedDict()
        for j in range(len(weights[element])):
            der_weights_square_error[element][
                j +
                1] = np.zeros(
                shape=np.shape(
                    weights[element][
                        j +
                        1]))

    der_scalings_square_error = OrderedDict()
    for element in elements:
        der_scalings_square_error[element] = OrderedDict()
    for element in elements:
        der_scalings_square_error[element]['intercept'] = 0.
        der_scalings_square_error[element]['slope'] = 0.

    for image_no in images.keys():

        o = {}
        D = {}
        delta = {}
        W = {}
        ohat = {}
        atoms = images[image_no]
        if sfp._save_fingerprint_derivatives is None:
            cf = CalculateFingerprints(sfp._cutoff, sfp._Gs, atoms)
        NN_energy = 0.
        real_energy = atoms.get_potential_energy(apply_constraint=False)

        for atom in atoms:
            index = atom.index
            symbol = atom.symbol
            indexfp = sfp.get_fp(image_no, index)
            o[index], ee = ff.get_output(indexfp, symbol)
            o[index][0] = indexfp
            NN_energy += ee
            N = len(o[index]) - 2
            weight = weights[symbol]
            D[index] = {}
            for k in range(1, N + 2):
                # calculating D matrix
                D[index][k] = np.zeros(
                    shape=(
                        np.size(
                            o[index][k]), np.size(
                            o[index][k])))
                for j in range(np.size(o[index][k])):
                    if activation == 'linear':  # linear
                        D[index][k][j, j] = 1.
                    elif activation == 'sigmoid':  # sigmoid
                        D[index][k][j, j] = float(
                            np.matrix(o[index][k])[0, j]) * float((1. - np.matrix(o[index][k])[0, j]))
                    elif activation == 'tanh':  # tanh
                        D[index][k][
                            j,
                            j] = float(
                            1. -
                            np.matrix(
                                o[index][k])[
                                0,
                                j] *
                            np.matrix(
                                o[index][k])[
                                0,
                                j])
            # Calculating delta
            delta[index] = {}
            # output layer
            delta[index][N + 1] = D[index][N + 1]
            # hidden layers
            W[index] = {}
            for j in range(len(weight)):
                W[index][j + 1] = np.delete(weight[j + 1], -1, 0)
            for k in range(N, 0, -1):
                delta[index][k] = np.dot(
                    D[index][k],
                    np.dot(
                        W[index][
                            k +
                            1],
                        delta[index][
                            k +
                            1]))
            # Calculating ohat
            ohat[index] = {}
            for k in range(1, N + 2):
                ohat[index][k - 1] = []
                for j in range(len(o[index][k - 1])):
                    ohat[index][k - 1].append(o[index][k - 1][j])
                ohat[index][k - 1].append(1.)

        square_error += (NN_energy - real_energy) ** 2.

        for atom in atoms:
            index = atom.index
            symbol = atom.symbol
            N = len(o[index]) - 2
            der_scalings_square_error[symbol][
                'intercept'] += 2. * (NN_energy - real_energy)
            der_scalings_square_error[symbol][
                'slope'] += 2. * (NN_energy - real_energy) * o[index][N + 1]
            for k in range(1, N + 2):
                der_weights_square_error[symbol][k] += 2. * float(NN_energy - real_energy) * float(
                    scalings[symbol]['slope']) * np.dot(np.matrix(ohat[index][k - 1]).T, np.matrix(delta[index][k]).T)

        if sfp._train_forces:

            real_forces = atoms.get_forces()
            nn_forces = np.zeros((len(atoms), 3))
            for self_atom in atoms:
                self_index = self_atom.index
                der_coordinates_D = {}
                der_coordinates_o = {}
                der_coordinates_delta = {}
                der_coordinates_ohat = {}
                der_coordinates_weights_atomic_output = {}
                for atom in atoms:
                    symbol = atom.symbol
                    index = atom.index
                    N = len(o[index]) - 2
                    weight = weights[symbol]
                    for i in range(3):
                        der_coordinates_o[(index, i)] = {}
                        if sfp._save_fingerprint_derivatives:
                            index_der_fp = sfp.get_der_fp(
                                image_no, (index, self_index, i))
                        else:
                            index_der_fp = cf.get_der_fingerprint(
                                symbol,
                                index,
                                self_index,
                                i)
                        der_coordinates_o[(index, i)][0] = index_der_fp
                        for k in range(1, N + 2):
                            # Calculating derivative of NN outputs w.r.t.
                            # coordinates
                            temp = np.dot(
                                np.matrix(der_coordinates_o[(index, i)][k - 1]), np.delete(weight[k], -1, 0))
                            der_coordinates_o[(index, i)][k] = []
                            for j in range(len(o[index][k])):
                                if activation == 'linear':  # linear function
                                    der_coordinates_o[(index, i)][k].append(
                                        float(temp[0, j]))
                                # sigmoid function
                                elif activation == 'sigmoid':
                                    der_coordinates_o[(index, i)][k].append(
                                        float(o[index][k][j] * (1. - o[index][k][j])) * float(temp[0, j]))
                                elif activation == 'tanh':  # tanh function
                                    der_coordinates_o[(index, i)][k].append(
                                        float(1. - o[index][k][j] * o[index][k][j]) * float(temp[0, j]))
                        der_coordinates_D[(index, i)] = {}
                        for k in range(1, N + 2):
                            # Calculating coordinate derivative of D matrix
                            der_coordinates_D[(index, i)][k] = np.zeros(
                                shape=(np.size(o[index][k]), np.size(o[index][k])))
                            for j in range(np.size(o[index][k])):
                                if activation == 'linear':  # linear
                                    der_coordinates_D[(index, i)][k][j, j] = 0.
                                elif activation == 'tanh':  # tanh
                                    der_coordinates_D[(index, i)][k][
                                        j, j] = - 2. * o[index][k][j] * der_coordinates_o[(index, i)][k][j]
                                elif activation == 'sigmoid':  # sigmoid
                                    der_coordinates_D[(index, i)][k][j, j] = der_coordinates_o[(index, i)][k][
                                        j] - 2. * o[index][k][j] * der_coordinates_o[(index, i)][k][j]
                        # Calculating coordinate derivative of delta
                        der_coordinates_delta[(index, i)] = {}
                        # output layer
                        der_coordinates_delta[
                            (index, i)][N + 1] = der_coordinates_D[(index, i)][N + 1]
                        # hidden layers
                        temp1 = {}
                        temp2 = {}
                        for k in range(N, 0, -1):
                            temp1[k] = np.dot(
                                W[index][
                                    k +
                                    1],
                                delta[index][
                                    k +
                                    1])
                            temp2[k] = np.dot(
                                W[index][k + 1], der_coordinates_delta[(index, i)][k + 1])
                            der_coordinates_delta[(index, i)][k] = np.dot(
                                der_coordinates_D[(index, i)][k], temp1[k]) + np.dot(D[index][k], temp2[k])
                        # Calculating coordinate derivative of ohat and
                        # coordinates weights derivative of atomic_output
                        der_coordinates_ohat[(index, i)] = {}
                        der_coordinates_weights_atomic_output[(index, i)] = {}
                        for k in range(1, N + 2):
                            der_coordinates_ohat[(index, i)][k - 1] = []
                            for j in range(len(der_coordinates_o[(index, i)][k - 1])):
                                der_coordinates_ohat[
                                    (index, i)][k - 1].append(der_coordinates_o[(index, i)][k - 1][j])
                            der_coordinates_ohat[(index, i)][k - 1].append(0.)
                            der_coordinates_weights_atomic_output[
                                (
                                    index,
                                    i)][k] = np.dot(
                                np.matrix(
                                    der_coordinates_ohat[
                                        (index,
                                         i)][
                                        k - 1]).T,
                                np.matrix(
                                    delta[index][k]).T) + np.dot(
                                np.matrix(
                                    ohat[index][
                                        k - 1]).T,
                                np.matrix(
                                    der_coordinates_delta[
                                        (index,
                                         i)][k]).T)

                        nn_forces[self_index][
                            i] += - float(scalings[symbol]['slope']) * der_coordinates_o[(index, i)][N + 1][0]

                for i in range(3):
                    square_error += (0.3333 * (
                        nn_forces[self_index][i] - real_forces[self_index][i]) ** 2.) / len(atoms)
                    for atom in atoms:
                        symbol = atom.symbol
                        index = atom.index
                        N = len(o[index]) - 2
                        for k in range(1, N + 2):
                            der_weights_square_error[symbol][k] += 0.66666667 * float(scalings[symbol]['slope']) * (- nn_forces[
                                self_index][i] + real_forces[self_index][i]) * der_coordinates_weights_atomic_output[(index, i)][k] / len(atoms)
                        der_scalings_square_error[symbol]['slope'] += 0.66666667 * der_coordinates_o[
                            (index, i)][N + 1][0] * (- nn_forces[self_index][i] + real_forces[self_index][i]) / len(atoms)

    queues.put(
        [square_error, der_weights_square_error, der_scalings_square_error])

##############################################################################


def compare_train_test_fingerprints(calculate_fingerprint, atoms,
                                    fingerprint_range):
    """function to compare train images with the test image and decide whether
    the prediction is interpolation or extrapolation.
    inputs:
        calculate_fingerprint: object of CalculateFingerprints class
        atoms: ASE atom object
        fingerprint_range: fingerprint range of the train images
    output:
        compare_train_test_fingerprints:
        integer: zero for interpolation, and one for extrapolation"""

    compare_train_test_fingerprints = 0

    for atom in atoms:
        index = atom.index
        symbol = atom.symbol
        indexfp = calculate_fingerprint.index_fingerprint(symbol, index)
        for i in range(len(indexfp)):
            if indexfp[i] < fingerprint_range[symbol][i][0] or \
                    indexfp[i] > fingerprint_range[symbol][i][1]:
                compare_train_test_fingerprints = 1
                break
    return compare_train_test_fingerprints

##############################################################################


class ExtrapolateError(Exception):

    """Kludge to raise error in the case of extrapolation"""
    pass

##############################################################################


def interpolate_images(images, json):
    """Function to remove extrapolation images from the "images" set"""

    json_file = paropen(json, 'rb')
    cutoff, activation, hiddenlayers, label, Gs, fingerprint_range, weights, \
        scalings = read_parameters_json(json_file)

    if isinstance(images, str):
        extension = os.path.splitext(images)[1]
        if extension == '.traj':
            images = io.PickleTrajectory(images, 'r')
        elif extension == '.db':
            images = io.read(images)

#   Images is converted to dictionary form
    dict_images = {}
    for i in range(len(images)):
        dict_images[i] = images[i]
    images = dict_images.copy()

#   Dictionary of interpolated images set is initialized
    interpolated_images = {}
    for image_no in images.keys():
        atoms = images[image_no]
        cf = CalculateFingerprints(cutoff, Gs, atoms)
        compare_train_test_fingerprints = 0
        for atom in atoms:
            index = atom.index
            symbol = atom.symbol
            indexfp = cf.index_fingerprint(symbol, index)
            for i in range(len(indexfp)):
                if indexfp[i] < fingerprint_range[symbol][i][0] or \
                        indexfp[i] > fingerprint_range[symbol][i][1]:
                    compare_train_test_fingerprints = 1
                    break
        if compare_train_test_fingerprints == 0:
            interpolated_images[image_no] = images[image_no]

    return interpolated_images
