#!/usr/bin/env python
"""
This module contains the CartesianNeural calculator, an ASE-like calculator
for implementing a basic neural network based on the Cartesian coordinates
of the atoms in the system. This will only work if the number and
identities of atoms in the systems under study do not change -- only the
atomic positions can change.

This module currently depends on an external library, "neurolab", which can
be found for example with "easy_install neurolab". Hopefully this
dependency will be removed in the future.
"""

import os
import numpy as np

from ase import io
from ase.calculators.calculator import Calculator, all_changes
from ase.parallel import paropen
import ffnet
from neural.utilities import ExtrapolateError, UntrainedError


class CartesianNeural(Calculator):
    """Sets up a neural network based on cartesian coordinates of the
    atoms. (As opposed to the Behler-Parinelllo method.) This method does
    not allow one to add or subtract atoms, but is useful for representing
    a local potential energy surface. All atoms must have identical
    numbering and unit cell.
    Here, extrapolate is defined by any cartesian coordinate (x, y, z) of
    an atom being in a region that is not covered by the scale argument. By
    default this is set to True to allow this to occur, but can be set to
    'truncate' to return the bounds instead of outliers or False to raise
    an exception when a user attempts to extrapolate.

    force_type can be one of 'analytical' (faster) or 'numerical' (can be
    used to check for bugs in analytical). This can be changed on-the-fly
    by modifying calc._force_type.

    label is the name of the file (or directory/file) to which to send
    output.
    """

    implemented_properties = ['energy', 'forces']

    def __init__(self, atoms=None, hiddenlayers=(5, 5), extrapolate=True,
                 force_type='analytical', label=None):
        #FIXME should eventually be able to re-establish itself by opening
        # a json file from a previous training.
        Calculator.__init__(self, atoms=atoms, label=label)
        self._network = None
        self._scale = None
        self._hiddenlayers = hiddenlayers
        self._extrapolate = extrapolate
        self._force_type = force_type

    def train(self, images, goal=0.01, maxtries=10, epochs=1000,
              callback=None):
        """Train the neural network to the atoms in the specified images or
        trajectory file. goal is the error goal (in MAE, eV).
        callback is the callback function called by scipy.fmin."""
        if type(images) == str:
            images = io.PickleTrajectory(images, 'r')
        self._scale = ScaleImages(images)
        scale = self._scale
        inputs, outputs = [], []
        for image in images:
            inputs.append(scale.get_scaled_positions(image.positions.copy(),
                          extrapolate=False))
            outputs.append([scale.get_scaled_energy(
                            image.get_potential_energy())])
        inputs = np.array(inputs)
        outputs = np.array(outputs)
        tries = 0
        last_error = np.inf
        while (tries < maxtries) and (last_error > goal):
            self.log('Training attempt #%i/%i' % (tries, maxtries))
            netsize = tuple([scale.positionlength]
                            + list(self._hiddenlayers) + [1])
            connections = ffnet.mlgraph(netsize)
            network = ffnet.ffnet(connections,
                                  lazy_derivative=False)
            self._network = network
            self._inputs = inputs  # for callback fxn
            self._outputs = outputs  # for callback fxn
            # FIXME : watch for 'lazy_derivative' keyword
            self.log('Training genetic.')
            network.train_genetic(input=inputs, target=outputs)
            predicts, _ = network.test(input=inputs,
                                       target=outputs, iprint=0)
            mae = check_error(scale.get_real_energy(outputs),
                              scale.get_real_energy(predicts))
            self.log('MAE: %.3f eV' % mae)
            self.log('Training BFGS.')
            #network.train_bfgs(input=inputs, target=outputs,
            #                   disp=0, factr=10., callback=callback)
            network.train_tnc(input=inputs, target=outputs,
                              callback=callback)
            predicts, _ = network.test(input=inputs,
                                       target=outputs, iprint=0)
            mae = check_error(scale.get_real_energy(outputs),
                              scale.get_real_energy(predicts))
            self.log('MAE: %.3f eV' % mae)
            last_error = mae
            tries += 1
        if last_error > goal:
            raise RuntimeError('Did not converge.')
        self._network = network
        self._scale = scale

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=all_changes):
        # FIXME : raise error if property other than positions changes
        Calculator.calculate(self, atoms, properties, system_changes)
        atoms = self.atoms
        scale = self._scale
        network = self._network
        scaledpositions = scale.get_scaled_positions(self.atoms.positions,
                                                     self._extrapolate)
        output = network.call([scaledpositions])[0][0]
        self.results['energy'] = scale.get_real_energy(output)
        if 'forces' in properties:
            if self._force_type == 'analytical':
                output = network.derivative(scaledpositions)[0]
                forces = scale.get_real_forces(output)
            elif self._force_type == 'numerical':
                newatoms = self.atoms.copy()
                newatoms.set_calculator(self)
                forces = self.calculate_numerical_forces(newatoms)
            else:
                raise AssertionError('Bad force specification.')

            self.results['forces'] = forces

    def log(self, message):
        """Writes message to file."""
        if not self.label:
            return
        f = paropen(os.path.join(self.directory, '%s.log' % self.prefix),
                    'a')
        f.write(message + '\n')
        f.close()


class ScaleImages:
    """Wrapper class to scale all image coordinates and images into the
    range -1 to +1 and to mask all dimensions that do not vary. That is, if
    certain atoms are fixed then their positions are masked.

    This will raise a warning if one tries to input something outside of
    the originally scaled range.

    Initialize with a list of images (ASE atoms objects). tol is the
    distance, in Angstroms, that if the whole range is less than it will
    treat it as invariant.
    """

    def __init__(self, images, tol=0.001):
        energies = []
        positions = np.zeros((len(images), len(images[0].positions.ravel())))
        for index, image in enumerate(images):
            energies.append(image.get_potential_energy())
            positions[index] = image.positions.ravel()
        self._energyrange = [min(energies), max(energies)]
        self._positionrange = []
        self.positionlength = 0
        for index in range(positions.shape[1]):
            width = max(positions[:, index]) - min(positions[:, index])
            if width < tol:
                self._positionrange.append([max(positions[:, index])
                                            - 0.5 * tol])  # invariant
            else:
                self._positionrange.append([min(positions[:, index]),
                                            max(positions[:, index])])
                self.positionlength += 1
        self._tol = tol

    def get_scaled_positions(self, positions, extrapolate):
        """Returns scaled, truncated, and unraveled version of a position
        list suitable to be sent to the neural network.
        If extrapolate is True, only returns values in the -1 to 1 range
        (sets outliers to either -1 or 1 depending on whether positive or
        negative).
        """
        positions = positions.ravel()
        scaled = []
        for coordinate, range in zip(positions, self._positionrange):
            if len(range) == 1:  # invariant
                if abs(coordinate - range[0]) > self._tol:
                    if not extrapolate:
                        raise UntrainedError('Trying to feed in '
                               'different value for coordinate which was '
                               ' invariant in training.')
            else:
                value = (-1. + 2. * (coordinate - range[0]) /
                              (range[1] - range[0]))
                if ((value > 1.) or (value < -1.)) and not extrapolate:
                    raise ExtrapolateError('Trying to extrapolate, which'
                                           ' is not allowed. Change to '
                                           'extrapolate=True if this is'
                                           ' desired.')
                if extrapolate == 'truncate':
                    if value > 1.:
                        value = 1.
                    elif value < -1.:
                        value = -1.
                scaled.append(value)
        return scaled

    def get_real_forces(self, scaledforces):
        """Returns the unscaled value of the forces (eV/A)."""
        realforces = np.zeros((len(self._positionrange) / 3, 3))
        xyz = 0
        atom_index = 0
        dE = -(self._energyrange[1] - self._energyrange[0])
        count = 0
        for range in self._positionrange:
            if len(range) == 1:  # invariant in training
                realforces[atom_index, xyz] = np.nan
            else:
                scaledforce = scaledforces[count]
                count += 1
                realforces[atom_index, xyz] = (dE / (range[1] - range[0])
                                               * scaledforce)
            xyz = (xyz + 1) % 3
            if xyz == 0:
                atom_index += 1
        return realforces

    def get_scaled_energy(self, energy):
        """Returns scaled energy suitable to be sent to the neural
        network."""
        return (-1. + 2. * (energy - self._energyrange[0]) /
                (self._energyrange[1] - self._energyrange[0]))

    def get_real_energy(self, scaled_energy):
        """Returns the actual energy to which the scaled value
        (scaled_energy) corresponds."""
        return (self._energyrange[0] + (scaled_energy + 1.) / 2. *
                (self._energyrange[1] - self._energyrange[0]))


def check_error(targets, predictions):
    """Calculates mean absolute error of the data set."""
    abs_err = 0.
    for target, prediction in zip(targets, predictions):
        abs_err += abs(prediction - target)
    return abs_err / len(targets)
