#!/usr/bin/env python
# SBATCH --time=20:30:00
# SBATCH --nodes=1
# SBATCH -J nn_C_Pt211
# SBATCH --ntasks-per-node=1
# SBATCH --partition=batch
# SBATCH --mem=4G

from matplotlib import pyplot
from ase.io.trajectory import PickleTrajectory
from neural.bp import BPNeural as Neural

images = PickleTrajectory('../C-Pt211.traj')

calc = Neural(label='C-Pt-tanh-0005')

##########################################################################
fig = pyplot.figure(figsize=(5., 10.))

# Case I: Energy.
ax = fig.add_subplot(311)

for i in range(len(images)):
    atoms = images[i]
    act_energy = atoms.get_potential_energy()
    atoms.set_calculator(calc)
    pred_energy = calc.get_potential_energy(atoms)
    print('%15.3f %15.3f %15.3f' %
          (act_energy, pred_energy, abs(act_energy - pred_energy)))
    ax.plot(act_energy, pred_energy, 'b.')

ax.set_xlabel('Actual energy, eV')
ax.set_ylabel('Predicted energy, eV')
ax.set_title('Sigmoid 50-5-5-1 NN Trained with Error 0.001')

##########################################################################

# Case II: Actual vs NN-Analytical Forces.
ax = fig.add_subplot(312)

atoms = images[-1]
act_force = atoms.get_forces()
atoms.set_calculator(calc)
nn_analytical_force = calc.get_forces(atoms)
for atom in atoms:
    index = atom.index
    for k in range(3):
        ax.plot(act_force[index][k], nn_analytical_force[index][k], 'b.')
        print('%15.3f %15.3f %15.3f' %
              (act_force[index][k], nn_analytical_force[index][k],
               abs(act_force[index][k] - nn_analytical_force[index][k])))

ax.set_xlabel('Actual forces, eV/Ang')
ax.set_ylabel('NN analytical forces, eV/Ang')
ax.set_title('Forces for the Last Image')

##########################################################################

# Case III: NN-Analytical and NN-Numerical Forces.
ax = fig.add_subplot(313)

nn_numerical_force = calc.calculate_numerical_forces(atoms, d=0.0001)
for atom in atoms:
    index = atom.index
    for k in range(3):
        ax.plot(nn_numerical_force[index][k],
                nn_analytical_force[index][k], 'b.')
        print('%15.3f %15.3f %15.3f' %
              (nn_numerical_force[index][k], nn_analytical_force[index][k],
               abs(nn_numerical_force[index][k] -
                   nn_analytical_force[index][k])))

ax.set_xlabel('NN-numerical forces')
ax.set_ylabel('NN-analytical forces')

fig.savefig('energy-force-plots-C-Pt211-last-image-sigmoid.png')

##########################################################################
# Sum of the forces in each direction:

fx = 0.
fy = 0.
fz = 0.

for atom in atoms:
    index = atom.index
    fx += nn_analytical_force[index][0]
    fy += nn_analytical_force[index][1]
    fz += nn_analytical_force[index][2]
print (fx, fy, fz)

fx = 0.
fy = 0.
fz = 0.
for atom in atoms:
    index = atom.index
    fx += nn_numerical_force[index][0]
    fy += nn_numerical_force[index][1]
    fz += nn_numerical_force[index][2]
print (fx, fy, fz)
##########################################################################
