#!/usr/bin/env python
"""Test of the cartesian neural network calculator. Randomly generates data
with the EMT potential in MD simulations. Both trains and tests getting
energy out of the calculator. Shows results for both interpolation and
extrapolation."""

import sys
sys.path.insert(0, '/home/aap/Dropbox/20140504-Neural')

from matplotlib import pyplot

from ase.calculators.emt import EMT
from ase.lattice.surface import fcc110
from ase import Atoms, Atom
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase import units
from ase.md import VelocityVerlet
from ase.constraints import FixAtoms

from neural.cartesian import CartesianNeural
from neural.utilities import randomize_images


def generate_data(count):
    """Generates test or training data with a simple MD simulation."""
    atoms = fcc110('Pt', (2, 2, 2), vacuum=7.)
    adsorbate = Atoms([Atom('Cu', atoms[7].position + (0., 0., 2.5)),
                       Atom('Cu', atoms[7].position + (0., 0., 5.))])
    atoms.extend(adsorbate)
    atoms.set_constraint(FixAtoms(indices=[0, 2]))
    atoms.set_calculator(EMT())
    MaxwellBoltzmannDistribution(atoms, 300. * units.kB)
    dyn = VelocityVerlet(atoms, dt=1. * units.fs)
    newatoms = atoms.copy()
    newatoms.set_calculator(EMT())
    newatoms.get_potential_energy()
    images = [newatoms]
    for step in range(count):
        dyn.run(5)
        newatoms = atoms.copy()
        newatoms.set_calculator(EMT())
        newatoms.get_potential_energy()
        images.append(newatoms)
    return images


fig = pyplot.figure(figsize=(5., 10.))

# Case I: Interpolation.
ax = fig.add_subplot(211)
print('Generating data.')
all_images = generate_data(120)
train_images, test_images = randomize_images(all_images)

print('Training network.')
calc = CartesianNeural()
calc.train(train_images, goal=0.01)

print('Testing network.')
for image in train_images:
    pred_energy = calc.get_potential_energy(atoms=image)
    act_energy = image.get_potential_energy()
    ax.plot(act_energy, pred_energy, 'b.')
for image in test_images:
    pred_energy = calc.get_potential_energy(atoms=image)
    act_energy = image.get_potential_energy()
    ax.plot(act_energy, pred_energy, 'r.')

ax.set_xlabel('Actual energy, eV')
ax.set_ylabel('Predicted energy, eV')
ax.set_title('Interpolate')


# Case II: Extrapolation
ax = fig.add_subplot(212)
print('Generating data.')
train_images = generate_data(100)
test_images = generate_data(100)

print('Training network.')
calc = CartesianNeural()
calc.train(train_images, goal=0.01)

print('Testing network.')
for image in train_images:
    pred_energy = calc.get_potential_energy(atoms=image)
    act_energy = image.get_potential_energy()
    ax.plot(act_energy, pred_energy, 'b.')
for image in test_images:
    pred_energy = calc.get_potential_energy(atoms=image)
    act_energy = image.get_potential_energy()
    ax.plot(act_energy, pred_energy, 'r.')

ax.set_xlabel('Actual energy, eV')
ax.set_ylabel('Predicted energy, eV')
ax.set_title('Extrapolate')

fig.savefig('cartesian-test.pdf')
