#!/usr/bin/env python
# SBATCH --time=50:00:00
# SBATCH --nodes=1
# SBATCH --ntasks-per-node=1
# SBATCH --partition=batch
# SBATCH --mem=4G

import sys
sys.path.insert(0, '/home/aap/Dropbox/20140504-Neural')
sys.path.insert(0, '/users/ap31/usr/svn/neural')
sys.path.insert(0, '/users/ap31/usr/svn/ase')

import sys
sys.path.insert(0, '/home/aap/Dropbox/20140504-Neural')

from matplotlib import pyplot
import numpy as np

from ase import Atoms, Atom
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md import VelocityVerlet
from ase import units

from neural.piggyback import Piggyback
from neural.utilities import ExtrapolateError

from neural import Neural
from ase.calculators.emt import EMT
from ase.lattice.surface import fcc110
from ase.constraints import FixAtoms


# Set up the piggyback calculator. This will not be used in the real MD
# because it has no forces. It will only be used to compare predictions in
# a dynamic environment.
real_calc = EMT()
piggybacks = [Neural(label='%i' % index)
              for index in range(5)]
piggybackcalc = Piggyback(real_calculator=real_calc, piggybacks=piggybacks,
                          label='piggyback', retrain_interval=50)


# Set up the atoms and attach an EMT calculator.
atoms = fcc110('Pt', (2, 2, 2), vacuum=7.)
adsorbate = Atoms([Atom('Cu', atoms[7].position + (0., 0., 2.5)),
                   Atom('Cu', atoms[7].position + (0., 0., 5.))])
atoms.extend(adsorbate)
atoms.set_constraint(FixAtoms(indices=[0, 2]))
calc = EMT()
atoms.set_calculator(calc)

# Initialize a figure to track the progress.
fig = pyplot.figure()
ax = fig.add_subplot(111)

# Run some MD.
MaxwellBoltzmannDistribution(atoms, 300. * units.kB)
energies = []
for step in range(1000):
    line = 'Step %i.' % step
    dyn = VelocityVerlet(atoms, dt=1. * units.fs)
    dyn.run(5)
    energies.append(atoms.get_potential_energy())
    if len(energies) > 1:
        ax.plot((step - 1, step), energies[-2:], '-', color='0.5')
    if piggybackcalc._trained:
        line += ' %12s' % 'Trained.'
    else:
        line += ' %12s' % 'Not trained.'
    if piggybackcalc._trained:
        pb_energies = []
        for subcalc in piggybackcalc._piggybacks:
            try:
                pb_energies.append(subcalc.get_potential_energy(atoms=atoms))
            except ExtrapolateError:
                pb_energies.append(np.nan)
        ax.plot([step] * len(pb_energies), pb_energies, 'kx')
        if np.isnan(pb_energies[0]):
            ax.text(step, energies[-1], 'e', color='blue', ha='center',
                    va='center')
    prediction = piggybackcalc.get_potential_energy(atoms=atoms)
    if piggybackcalc.predicted:
        line += ' %14s' % 'Predicted.'
    else:
        line += ' %14s' % 'Not predicted.'
    print(line)
    if piggybackcalc.predicted:
        ax.plot(step, prediction, 'or')

    fig.savefig('out.pdf')
