#!/usr/bin/env python
# SBATCH --time=01:30:00
# SBATCH --nodes=1
# SBATCH -J MD_multiple_training
# SBATCH --ntasks-per-node=64
# SBATCH --partition=batch
# SBATCH --mem=64G

"""Images are randomly generates data with MD simulations. Train set is
generated by sequentially adding images. At each iteration, a set of 1000
images is generated as the test set.  The RMSE of the test set
is calculated first, and then the test set is added to the train set."""

##############################################################################

from neural.bp import BPNeural as Neural
from neural.bp import read_parameters_json, make_symmetry_functions, FeedForward
from neural.bp import CalculateFingerprints, calculate_error_per_image_atom
from neural.MD_generate_images import generate_data
from ase.parallel import paropen
import time
from collections import OrderedDict

##############################################################################
weights = None
scalings = None
hiddenlayers = {'Cu': (5, 5), 'Pt': (5, 5)}
elements = sorted(hiddenlayers.keys())
Gs = make_symmetry_functions(elements, None)

train_images = generate_data(99)
test_images = []

epochlog = paropen('Epochs-log', 'w')
head = ('%7s ' ' %25s' ' %15s' '%15s\n')
epochlog.write(head % ('Epoch', 'Time', 'Error, eV^2', 'RMSE, meV/atom'))
epochlog.write(head % ('=' * 6, '=' * 24, '=' * 14, '=' * 14))

line = ('%7s:' ' %25s' ' %15.2f' '%15.2f\n')

for i in range(100):

    if i > 0:
        train_images = [train_images[m] for m in range(len(train_images))] + \
            [test_images[n] for n in range(len(test_images))]

    try:
        fp = open('%i-trained-parameters-log.json' % (i - 1), "rb")
        cutoff, activation, hiddenlayers, label, Gs, weights, scalings = \
            read_parameters_json(fp)
    except:
        fp = None

    calc = Neural(
        hiddenlayers=hiddenlayers,
        label='%i' %
        i,
        weights=weights,
        scalings=scalings,
        Gs=Gs,
        activation='tanh',
        fp_label='%i' %
        (i -
         1))

    calc.train(train_images, converge=0.005)

    ##########################################################################

    fp = open('%i-trained-parameters-log.json' % i, 'rb')
    cutoff, activation, hiddenlayers, label, Gs, weights, scalings =\
        read_parameters_json(fp)

    ff = FeedForward(hiddenlayers, weights, scalings, activation)

    ##########################################################################

    test_images = generate_data(99)
    test_images_dict = {}
    for i in range(len(test_images)):
        test_images_dict[i] = test_images[i]
    sfp = {}
    for image_no in test_images_dict.keys():
        image = test_images_dict[image_no]
        cf = CalculateFingerprints(cutoff=cutoff, Gs=Gs, atoms=image)
        for atom in image:
            index = atom.index
            symbol = atom.symbol
            indexfp = cf.index_fingerprint(symbol, index)
            sfp[(image_no, index)] = indexfp

    outputs = OrderedDict()
    NN_energies = OrderedDict()
    for image_no in test_images_dict.keys():
        NN_energy = 0.
        o = OrderedDict()
        for atom in test_images_dict[image_no]:
            index = atom.index
            symbol = atom.symbol
            indexfp = sfp[(image_no, index)]
            o[index], ee = ff.get_output(indexfp, symbol)
            NN_energy += ee
        outputs[image_no] = o
        NN_energies[image_no] = NN_energy

    rmse_per_atom, square_error = \
        calculate_error_per_image_atom(test_images_dict, NN_energies)

    line = ('%7s:' ' %25s' ' %15.2f' '%15.2f\n')
    epochlog.write(line % (i, time.ctime(), square_error,
                           rmse_per_atom * 1000.))

epochlog.close()

##############################################################################
##############################################################################
##############################################################################
